#ifndef SOAP_ON
#define SOAP_ON

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/queue.h>

#include "boa.h"

typedef char s8;
typedef unsigned char u8;

typedef short s16;
typedef unsigned short u16;

typedef int s32;
typedef unsigned int u32;

typedef unsigned long long u64;


#define MD5_DIGEST_SIZE			16
#define MD5_HMAC_BLOCK_SIZE		64
#define MD5_BLOCK_WORDS		16
#define MD5_HASH_WORDS		4

struct md5_ctx 
{
	u32 hash[MD5_HASH_WORDS];
	u32 block[MD5_BLOCK_WORDS];

	u32 byte_count; /* 32 bit length is enough for currently using. */
};

enum {
	CONFIG_STARTED = 1,
	CONFIG_FINISHED
};

 struct method_struct
{
	const char *methodName; 
	void (*methodImpl)(request *);
};

 struct portmap
{
	int enabled;

	char protocol[8];
	char intport[8];
	char extport[8];
	char inthost[32];
	char desc[65];

	struct portmap *prev;
	struct portmap *next;
} ;

static struct portmap *portmap_head = NULL;
static struct portmap *portmap_tail = NULL;

#define IS_WHITE_SPACE(c) ((c==' ') || (c=='\t') || (c=='\r') || (c=='\n'))

struct xmlparser {
	char *xmlstart;
	char *xmlend;
	char *xml;	/* pointer to current character */
	int xmlsize;
	void *data;
	void (*starteltfunc) (void *, char *, int);
	void (*endeltfunc) (void *, const char *, int);
	void (*datafunc) (void *, char *, int);
	void (*attfunc) (void *, const char *, int, const char *, int);
};

struct NameValue {
	LIST_ENTRY(NameValue) entries;
	char *name;
	char *value;
};

struct NameValueParserData {
	LIST_HEAD(listhead, NameValue) head;
	char *curelt;
};

void parsexml(struct xmlparser *);

void ParseNameValue(char *buffer, int bufsize,
               struct NameValueParserData *data);

void ClearNameValueList(struct NameValueParserData *pdata);

char *GetValueFromNameValueList(struct NameValueParserData *pdata,
                          const char *Name);

void SoapError(request *req, char *err);

void md5_init(void *ctx);
void md5_update(void *ctx, const u8 *data, unsigned int len);
void md5_final(void *ctx, u8 *out);
void wep_key_64(char *phrase, char wepkey[4][32]);
char * wep_key_128(char *phrase);

#endif

