#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/init.h>
#include <linux/resource.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>
#include <asm/types.h>
#include <asm/irq.h>
#include <asm/delay.h>

#include "ar7100.h"

#include <linux/fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/proc_fs.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/raw.h>
#include <linux/capability.h>
#include <linux/uio.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/sysctl.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/mach-ar7100/ar9100.h>
#include <linux/kdev_t.h>

#define AR7100_FACTORY_RESET           0x89ABCDEF
#define AR7100_HANDLE_KASSERT       (AR7100_FACTORY_RESET+1)
#define REASON_RESET 0x1
#define REASON_KASSERT 0x2
static atomic_t ar7100_fr_status = ATOMIC_INIT(0);
static volatile int ar7100_fr_opened=0;
static wait_queue_head_t ar7100_fr_wq;
static unsigned char wakeup_reason = 0;
 /*
  * By default KASSERT is enabled
  * Means KASSERT shall cause panic
  */
static unsigned char enable_kassert = 1;

static void ar7100_misc_wakeup_process(void)
{
          wakeup_reason = REASON_KASSERT;
          wake_up(&ar7100_fr_wq);
}

void ar7xxx_misc_wakeup_process(void)
{
        ar7100_misc_wakeup_process();
}
EXPORT_SYMBOL(ar7xxx_misc_wakeup_process);

void ar7xxx_set_kassert(unsigned char enable)
{
        enable_kassert = enable;
}
EXPORT_SYMBOL(ar7xxx_set_kassert);

unsigned char ar7xxx_is_kassert_enabled(void)
{
        return enable_kassert;
}
EXPORT_SYMBOL(ar7xxx_is_kassert_enabled);


/*
 * GPIO interrupt stuff
 */
typedef enum {
    INT_TYPE_EDGE,
    INT_TYPE_LEVEL,
}ar7100_gpio_int_type_t;

typedef enum {
    INT_POL_ACTIVE_LOW,
    INT_POL_ACTIVE_HIGH,
}ar7100_gpio_int_pol_t;

struct gpio_priv_t {
	char data_buffer[2];
	int data_available;
	unsigned long prevtime;
	unsigned int last_edge;
	unsigned int interval;
	wait_queue_head_t read_queue;
};



/* 
** Simple Config stuff
*/

#if !defined(IRQ_NONE)
#define IRQ_NONE
#define IRQ_HANDLED
#endif /* !defined(IRQ_NONE) */

#ifdef CONFIG_AR9100
#define JUMPSTART_GPIO    8
#else
#define JUMPSTART_GPIO    3
#ifdef CONFIG_CUS100
#define JUMPSTART_GPIO    7
#endif
#endif

typedef irqreturn_t(*sc_callback_t)(int, void *, struct pt_regs *);

/*
** Multiple Simple Config callback support
** For multiple radio scenarios, we need to post the button push to
** all radios at the same time.  However, there is only 1 button, so
** we only have one set of GPIO callback pointers.
**
** Creating a structure that contains each callback, tagged with the
** name of the device registering the callback.  The unregister routine
** will need to determine which element to "unregister", so the device
** name will have to be passed to unregister also
*/

typedef struct {
char   *name;
sc_callback_t  registered_cb;
void                   *cb_arg;
} Multi_Callback_t;

/*
** Specific instance of the callback structure
*/

static Multi_Callback_t        SCcallback[2] = {{NULL,NULL,NULL},{NULL,NULL,NULL}};

/*
** These are for the GPIO, so we don't create multiple instances
*/

static int ignore_pushbutton = 0;
static struct proc_dir_entry *simple_config_entry = NULL;
static struct proc_dir_entry *simulate_push_button_entry = NULL;
static struct proc_dir_entry *tricolor_led_entry = NULL;

static struct proc_dir_entry *wps_entry = NULL;

void ar7100_gpio_config_int(int gpio, 
                       ar7100_gpio_int_type_t type,
                       ar7100_gpio_int_pol_t polarity)
{
    u32 val;

    /*
     * allow edge sensitive/rising edge too
     */
    if (type == INT_TYPE_LEVEL) {
        /* level sensitive */
        ar7100_reg_rmw_set(AR7100_GPIO_INT_TYPE, (1 << gpio));
    }
    else {
       /* edge triggered */
       val = ar7100_reg_rd(AR7100_GPIO_INT_TYPE);
       val &= ~(1 << gpio);
       ar7100_reg_wr(AR7100_GPIO_INT_TYPE, val);
    }

    if (polarity == INT_POL_ACTIVE_HIGH) {
        ar7100_reg_rmw_set (AR7100_GPIO_INT_POLARITY, (1 << gpio));
    }
    else {
       val = ar7100_reg_rd(AR7100_GPIO_INT_POLARITY);
       val &= ~(1 << gpio);
       ar7100_reg_wr(AR7100_GPIO_INT_POLARITY, val);
    }

    ar7100_reg_rmw_set(AR7100_GPIO_INT_ENABLE, (1 << gpio));
}

void
ar7100_gpio_config_output(int gpio)
{
    ar7100_reg_rmw_set(AR7100_GPIO_OE, (1 << gpio));
}

void
ar7100_gpio_config_input(int gpio)
{
    ar7100_reg_rmw_clear(AR7100_GPIO_OE, (1 << gpio));
}

void
ar7100_gpio_out_val(int gpio, int val)
{
    if (val & 0x1) {
        ar7100_reg_rmw_set(AR7100_GPIO_OUT, (1 << gpio));
    }
    else {
        ar7100_reg_rmw_clear(AR7100_GPIO_OUT, (1 << gpio));
    }
}

int
ar7100_gpio_in_val(int gpio)
{
    return((1 << gpio) & (ar7100_reg_rd(AR7100_GPIO_IN)));
}

static void
ar7100_gpio_intr_enable(unsigned int irq)
{
    ar7100_reg_rmw_set(AR7100_GPIO_INT_MASK, 
                      (1 << (irq - AR7100_GPIO_IRQ_BASE)));
}

static void
ar7100_gpio_intr_disable(unsigned int irq)
{
    ar7100_reg_rmw_clear(AR7100_GPIO_INT_MASK, 
                        (1 << (irq - AR7100_GPIO_IRQ_BASE)));
}

static unsigned int
ar7100_gpio_intr_startup(unsigned int irq)
{
	ar7100_gpio_intr_enable(irq);
	return 0;
}

static void
ar7100_gpio_intr_shutdown(unsigned int irq)
{
	ar7100_gpio_intr_disable(irq);
}

static void
ar7100_gpio_intr_ack(unsigned int irq)
{
	ar7100_gpio_intr_disable(irq);
}

static void
ar7100_gpio_intr_end(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED | IRQ_INPROGRESS)))
		ar7100_gpio_intr_enable(irq);
}

static void
ar7100_gpio_intr_set_affinity(unsigned int irq, unsigned long mask)
{
	/* 
     * Only 1 CPU; ignore affinity request
     */
}

struct hw_interrupt_type ar7100_gpio_intr_controller = {
	"AR7100 GPIO",
	ar7100_gpio_intr_startup,
	ar7100_gpio_intr_shutdown,
	ar7100_gpio_intr_enable,
	ar7100_gpio_intr_disable,
	ar7100_gpio_intr_ack,
	ar7100_gpio_intr_end,
	ar7100_gpio_intr_set_affinity,
};

void
ar7100_gpio_irq_init(int irq_base)
{
	int i;

	for (i = irq_base; i < irq_base + AR7100_GPIO_IRQ_COUNT; i++) {
		irq_desc[i].status  = IRQ_DISABLED;
		irq_desc[i].action  = NULL;
		irq_desc[i].depth   = 1;
		irq_desc[i].handler = &ar7100_gpio_intr_controller;
	}
}


void register_simple_config_callback (char *cbname, void *callback, void *arg)
{
        printk("!!!!! SC Callback Registration for %s\n",cbname);
        if(!SCcallback[0].name)
        {
                SCcallback[0].name = cbname;
                SCcallback[0].registered_cb = (sc_callback_t) callback;
                SCcallback[0].cb_arg = arg;
        }
        else
        {
                SCcallback[1].name = cbname;
                SCcallback[1].registered_cb = (sc_callback_t) callback;
                SCcallback[1].cb_arg = arg;
        }
}
EXPORT_SYMBOL(register_simple_config_callback);

void unregister_simple_config_callback (char *cbname)
{
        if(SCcallback[1].name == cbname)
        {
                SCcallback[1].name = NULL;
                SCcallback[1].registered_cb = NULL;
                SCcallback[1].cb_arg = NULL;
        }
        else if(SCcallback[0].name == cbname)
        {
                SCcallback[0].name = NULL;
                SCcallback[0].registered_cb = NULL;
                SCcallback[0].cb_arg = NULL;
        }
        else
        {
                printk("!&!&!&!& ERROR: Unknown callback name %s\n",cbname);
        }
}
EXPORT_SYMBOL(unregister_simple_config_callback);

/*
 * Irq for front panel SW jumpstart switch
 * Connected to XSCALE through GPIO4
 */
irqreturn_t franky_jumpstart_irq(int cpl, void *dev_id, struct pt_regs *regs)
{
	printk("my irq##########################################\n");
}


irqreturn_t jumpstart_irq(int cpl, void *dev_id, struct pt_regs *regs)
{
    printk(".........");
    if (ignore_pushbutton) {
        ar7100_gpio_config_int (JUMPSTART_GPIO,INT_TYPE_LEVEL,
                                INT_POL_ACTIVE_HIGH);
        ignore_pushbutton = 0;
        return IRQ_HANDLED;
    }

    ar7100_gpio_config_int (JUMPSTART_GPIO,INT_TYPE_LEVEL,INT_POL_ACTIVE_LOW);
    ignore_pushbutton = 1;

    printk ("calling simple_config callback..\n");

    if (SCcallback[0].registered_cb) {
        SCcallback[0].registered_cb (cpl, SCcallback[0].cb_arg, regs);
    }
#if 0
    if (SCcallback[1].registered_cb) {
        SCcallback[1].registered_cb (cpl, SCcallback[1].cb_arg, regs);
    }
#endif
    return IRQ_HANDLED;
}

static int push_button_read (char *page, char **start, off_t off,
                               int count, int *eof, void *data)
{
    return 0;
}

static int push_button_write (struct file *file, const char *buf,
                                        unsigned long count, void *data)
{
   if (0 == strncmp(buf,"pbc",3) || 0 == strncmp(buf,"sta_pbc",7)) {
       if (SCcallback[0].registered_cb)
           SCcallback[0].registered_cb (0, SCcallback[0].cb_arg, (struct pt_regs *)buf);
   }
    return count;
}

static int wps_read (char *page, char **start, off_t off,
                              int count, int *eof, void *data)
{
    if (SCcallback[0].registered_cb)
        SCcallback[0].registered_cb (0, SCcallback[0].cb_arg, (struct pt_regs *)"info");
    if (SCcallback[1].registered_cb)
        SCcallback[1].registered_cb (0, SCcallback[1].cb_arg, (struct pt_regs *)"info");
    return 0;
}

static int wps_write (struct file *file, const char *buf,
                        unsigned long count, void *data)
{
    if (0 == strncmp(buf,"pin=",4) || 0 == strncmp(buf,"stop",4)) {
        if (SCcallback[0].registered_cb)
            SCcallback[0].registered_cb (0, SCcallback[0].cb_arg, (struct pt_regs *)buf);
#if 0
	if (SCcallback[1].registered_cb)
            SCcallback[1].registered_cb (0, SCcallback[1].cb_arg, (struct pt_regs *)buf);
#endif
       }
     return count;
 }


#define TRICOLOR_LED_GREEN_PIN  4         /* GPIO 4 */
#define TRICOLOR_LED_YELLOW_PIN 4         /* GPIO 4 */
#define OFF 1
#define ON 0

typedef enum {
        LED_STATE_OFF   =       0,
        LED_STATE_GREEN =       1,
        LED_STATE_YELLOW =      2,
        LED_STATE_ORANGE =      3,
        LED_STATE_MAX =         4
} led_state_e;

static led_state_e gpio_tricolorled = LED_STATE_OFF;

static int gpio_tricolor_led_read (char *page, char **start, off_t off,
               int count, int *eof, void *data)
{
    return sprintf (page, "%d\n", gpio_tricolorled);
}

static int gpio_tricolor_led_write (struct file *file, const char *buf,
                                        unsigned long count, void *data)
{
    u_int32_t val, green_led_onoff = 0, yellow_led_onoff = 0;

    if (sscanf(buf, "%d", &val) != 1)
        return -EINVAL;

    if (val >= LED_STATE_MAX)
        return -EINVAL;

    if (val == gpio_tricolorled)
        return count;

    switch (val) {
        case LED_STATE_OFF :
                green_led_onoff = OFF;   /* both LEDs OFF */
                yellow_led_onoff = OFF;
                break;

        case LED_STATE_GREEN:
                green_led_onoff = ON;    /* green ON, Yellow OFF */
                yellow_led_onoff = OFF;
                break;

        case LED_STATE_YELLOW:
                green_led_onoff = OFF;   /* green OFF, Yellow ON */
                yellow_led_onoff = ON;
                break;

        case LED_STATE_ORANGE:
                green_led_onoff = ON;    /* both LEDs ON */
                yellow_led_onoff = ON;
                break;
    }

    ar7100_gpio_out_val (TRICOLOR_LED_GREEN_PIN, green_led_onoff);
   // ar7100_gpio_out_val (TRICOLOR_LED_YELLOW_PIN, yellow_led_onoff);
    gpio_tricolorled = val;

    return count;
}


static int create_simple_config_led_proc_entry (void)
{
    if (simple_config_entry != NULL) {
        printk ("Already have a proc entry for /proc/simple_config!\n");
        return -ENOENT;
    }

    simple_config_entry = proc_mkdir("simple_config", NULL);
    if (!simple_config_entry)
        return -ENOENT;

    simulate_push_button_entry = create_proc_entry ("push_button", 0644,
                                                      simple_config_entry);
    if (!simulate_push_button_entry)
        return -ENOENT;

    simulate_push_button_entry->write_proc = push_button_write;
    simulate_push_button_entry->read_proc = push_button_read;

    tricolor_led_entry = create_proc_entry ("tricolor_led", 0644,
                                            simple_config_entry);
    if (!tricolor_led_entry)
        return -ENOENT;

    tricolor_led_entry->write_proc = gpio_tricolor_led_write;
    tricolor_led_entry->read_proc = gpio_tricolor_led_read;

    wps_entry = create_proc_entry ("wps", 0644,
                                    simple_config_entry);
    if (!wps_entry)
        return -ENOENT;

    wps_entry->write_proc = wps_write;
    wps_entry->read_proc = wps_read;


    /* configure gpio as outputs */
    ar7100_gpio_config_output (TRICOLOR_LED_GREEN_PIN); 
    //ar7100_gpio_config_output (TRICOLOR_LED_YELLOW_PIN); 

    /* switch off the led */
    ar7100_gpio_out_val(TRICOLOR_LED_GREEN_PIN, OFF);
    //ar7100_gpio_out_val(TRICOLOR_LED_YELLOW_PIN, OFF);

    return 0;
}



int __init pbXX_simple_config_init(void)
{
#ifdef CONFIG_CUS100
	u32 mask = 0;
#endif
    int req;
    int i;
   /*
    struct gpio_priv_t *priv;
    if(!(priv = kmalloc(sizeof(struct gpio_priv_t), GFP_KERNEL)))
	    printk("kmalloc failed\n");
    priv->data_available = 0;
    priv->prevtime = 0;
    priv->last_edge = LOW;
    priv->interval = 0;
    init_waitqueue_head(&priv->read_queue);

*/







    
#ifdef CONFIG_CUS100
	mask = ar7100_reg_rd(AR7100_MISC_INT_MASK);
	ar7100_reg_wr(AR7100_MISC_INT_MASK, mask | (1 << 2)); /* Enable GPIO interrupt mask */
    ar7100_gpio_config_int (JUMPSTART_GPIO, INT_TYPE_LEVEL,INT_POL_ACTIVE_HIGH);
	ar7100_gpio_intr_enable(JUMPSTART_GPIO);
	ar7100_gpio_config_input(JUMPSTART_GPIO);
#else
    /* configure GPIO 3 as level triggered interrupt */ 
    ar7100_gpio_config_int (JUMPSTART_GPIO, INT_TYPE_LEVEL,INT_POL_ACTIVE_HIGH);




	// just test by franky
/*
	ar7100_gpio_intr_enable(AR7100_GPIO_IRQn(21));
	ar7100_gpio_config_input(21);
	ar7100_gpio_config_int (21, INT_TYPE_EDGE,INT_POL_ACTIVE_LOW);
*/
#endif


    req = request_irq (AR7100_GPIO_IRQn(JUMPSTART_GPIO), jumpstart_irq, 0,
    "SW JUMPSTART", NULL);
  /*
       	req = request_irq (AR7100_GPIO_IRQn(21), franky_jumpstart_irq, 0,
		    "ar91xxgpiointr", NULL);
   */

   //test by franky end 
  
    if (req != 0) {
        printk (KERN_ERR "********************************************************unable to request IRQ for SWJUMPSTART GPIO (error %d)\n", req);
    }

    create_simple_config_led_proc_entry ();
    return 0;
}

subsys_initcall(pbXX_simple_config_init);
